from enum import auto
from typing import List
from typing import Optional

from gw.enums import GwStrEnum


class RecognizedCurrencyUnit(GwStrEnum):
    """
    Unit of currency

    Enum recognized.currency.unit version 000 in the GridWorks Type registry.

    Used by used by multiple Application Shared Languages (ASLs). For more information:
      - [ASLs](https://gridworks-type-registry.readthedocs.io/en/latest/)
      - [Global Authority](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#recognizedcurrencyunit)

    Values (with symbols in parens):
      - Unknown (00000000)
      - USD (e57c5143): US Dollar
      - GBP (f7b38fc5): Pounds sterling
    """

    UNKNOWN = auto()
    USD = auto()
    GBP = auto()

    @classmethod
    def default(cls) -> "RecognizedCurrencyUnit":
        """
        Returns default value (in this case UNKNOWN)
        """
        return cls.UNKNOWN

    @classmethod
    def values(cls) -> List[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]

    @classmethod
    def version(cls, value: Optional[str] = None) -> str:
        """
        Returns the version of the class (default) used by this package or the
        version of a candidate enum value (always less than or equal to the version
        of the class)

        Args:
            value (Optional[str]): None (for version of the Enum itself) or
            the candidate enum value.

        Raises:
            ValueError: If the value is not one of the enum values.

        Returns:
            str: The version of the enum used by this code (if given no
            value) OR the earliest version of the enum containing the value.
        """
        if value is None:
            return "000"
        if not isinstance(value, str):
            raise ValueError(f"This method applies to strings, not enums")
        if value not in value_to_version.keys():
            raise ValueError(f"Unknown enum value: {value}")
        return value_to_version[value]

    @classmethod
    def enum_name(cls) -> str:
        """
        The name in the GridWorks Type Registry (recognized.currency.unit)
        """
        return "recognized.currency.unit"

    @classmethod
    def enum_version(cls) -> str:
        """
        The version in the GridWorks Type Registry (000)
        """
        return "000"

    @classmethod
    def symbol_to_value(cls, symbol: str) -> str:
        """
        Given the symbol sent in a serialized message, returns the encoded enum.

        Args:
            symbol (str): The candidate symbol.

        Returns:
            str: The encoded value associated to that symbol. If the symbol is not
            recognized - which could happen if the actor making the symbol is using
            a later version of this enum, returns the default value of "Unknown".
        """
        if symbol not in symbol_to_value.keys():
            return cls.default().value
        return symbol_to_value[symbol]

    @classmethod
    def value_to_symbol(cls, value: str) -> str:
        """
        Provides the encoding symbol for a RecognizedCurrencyUnit enum to send in seriliazed messages.

        Args:
            symbol (str): The candidate value.

        Returns:
            str: The symbol encoding that value. If the value is not recognized -
            which could happen if the actor making the message used a later version
            of this enum than the actor decoding the message, returns the default
            symbol of "00000000".
        """
        if value not in value_to_symbol.keys():
            return value_to_symbol[cls.default().value]
        return value_to_symbol[value]

    @classmethod
    def symbols(cls) -> List[str]:
        """
        Returns a list of the enum symbols
        """
        return [
            "00000000",
            "e57c5143",
            "f7b38fc5",
        ]


symbol_to_value = {
    "00000000": "UNKNOWN",
    "e57c5143": "USD",
    "f7b38fc5": "GBP",
}

value_to_symbol = {value: key for key, value in symbol_to_value.items()}

value_to_version = {
    "UNKNOWN": "000",
    "USD": "000",
    "GBP": "000",
}



from typing import List

from gw.data_classes.market_type import MarketType
from gw.enums import MarketTypeName


def check_is_market_name(v: str) -> None:
    try:
        x = v.split(".")
    except AttributeError:
        raise ValueError(f"{v} failed to split on '.'")
    if not x[0] in MarketTypeName.values():
        raise ValueError(f"{v} not recognized MarketType")
    g_node_alias = ".".join(x[1:])
    check_is_left_right_dot(g_node_alias)


def check_is_market_slot_name_lrd_format(v: str) -> None:
    """
    MaketSlotNameLrdFormat: the format of a MarketSlotName.
      - The first word must be a MarketTypeName
      - The last word (unix time of market slot start) must
      be a 10-digit integer divisible by 300 (i.e. all MarketSlots
      start at the top of 5 minutes)
      - More strictly, the last word must be the start of a
      MarketSlot for that MarketType (i.e. divisible by 3600
      for hourly markets)
      - The middle words have LeftRightDot format (GNodeAlias
      of the MarketMaker)
    Example: rt60gate5.d1.isone.ver.keene.1673539200

    """
    try:
        x = v.split(".")
    except AttributeError:
        raise ValueError(f"{v} failed to split on '.'")
    slot_start = x[-1]
    try:
        slot_start = int(slot_start)
    except ValueError:
        raise ValueError(f"slot start {slot_start} not an int")
    check_is_reasonable_unix_time_s(slot_start)
    if slot_start % 300 != 0:
        raise ValueError(f"slot start {slot_start} not a multiple of 300")

    market_name = ".".join(x[:-1])
    try:
        check_is_market_name(market_name)
    except ValueError as e:
        raise ValueError(f"e")
    market_type = MarketType.by_id[MarketTypeName(x[0])]
    if not slot_start % (market_type.duration_minutes * 60) == 0:
        raise ValueError(
            f"market_slot_start_s mod {(market_type.duration_minutes * 60)} must be 0"
        )


def check_is_left_right_dot(candidate: str) -> None:
    """Lowercase AlphanumericStrings separated by dots (i.e. periods), with most
    significant word to the left.  I.e. `d1.ne` is the child of `d1`.
    Checking the format cannot verify the significance of words. All
    words must be alphanumeric. Most significant word must start with
    an alphabet charecter


    Raises:
        ValueError: if candidate is not of lrd format (e.g. d1.iso.me.apple)
    """
    try:
        x: List[str] = candidate.split(".")
    except:
        raise ValueError("Failed to seperate into words with split'.'")
    first_word = x[0]
    first_char = first_word[0]
    if not first_char.isalpha():
        raise ValueError(
            f"Most significant word must start with alphabet char. Got '{first_word}'"
        )
    for word in x:
        if not word.isalnum():
            raise ValueError(
                f"words seperated by dots must be alphanumeric. Got '{word}'"
            )
    if not candidate.islower():
        raise ValueError(f"alias must be lowercase. Got '{candidate}'")


def check_is_reasonable_unix_time_s(v: int) -> None:
    """
    ReasonableUnixTimeS format: time in unix seconds between Jan 1 2000 and Jan 1 3000

    Raises:
        ValueError: if not ReasonableUnixTimeS format
    """
    from datetime import datetime
    from datetime import timezone

    start_date = datetime(2000, 1, 1, tzinfo=timezone.utc)
    end_date = datetime(3000, 1, 1, tzinfo=timezone.utc)

    start_timestamp = int(start_date.timestamp())
    end_timestamp = int(end_date.timestamp())

    if v < start_timestamp:
        raise ValueError(f"{v} must be after Jan 1 2000")
    if v > end_timestamp:
        raise ValueError(f"{v} must be before Jan 1 3000")



from gw.property_format import check_is_market_name
from gw.property_format import check_is_market_slot_name_lrd_format


__all__ = [
    "check_is_market_name",
    "check_is_market_slot_name_lrd_format",
]



from gw_test.wait import AwaitablePredicate
from gw_test.wait import ErrorStringFunction
from gw_test.wait import Predicate
from gw_test.wait import StopWatch
from gw_test.wait import await_for
from gw_test.wait import wait_for


__all__ = [
    "AwaitablePredicate",
    "ErrorStringFunction",
    "Predicate",
    "StopWatch",
    "await_for",
    "wait_for",
]



class GwTypeError(Exception):
    """Base class for errors in making GridWorks Types"""

    pass


class BlockChainError(Exception):
    """Base class for errors related to Blockchain"""

    pass


class DcError(Exception):
    """Base class for dataclass errors"""

    pass


class RegistryError(Exception):
    """Base class for registry errors"""

    pass



import enum


class GwStrEnum(str, enum.Enum):
    """
    Mimics fastapi-utils use of StrEnum, which diverges from the
    python-native StrEnum for python 3.11+.  Specifically if

    class Foo(GwStrEnum):
        Bar = auto()

    then

    Foo.Bar.value is 'Bar'

    """

    def _generate_next_value_(name, start, count, last_values):
        return name



from enum import auto
from typing import List
from typing import Optional

from gw.enums import GwStrEnum


class MessageCategory(GwStrEnum):
    """
    Categorizes how GridWorks messages are sent and decoded/encoded

    Enum message.category version 000 in the GridWorks Type registry.

    Used by used by multiple Application Shared Languages (ASLs). For more information:
      - [ASLs](https://gridworks-type-registry.readthedocs.io/en/latest/)
      - [Global Authority](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#messagecategory)

    Values (with symbols in parens):
      - Unknown (00000000): Default value
      - RabbitJsonDirect (9057ab24): Serialized Json message sent on the world rabbit broker from one GNode
        actor to another
      - RabbitJsonBroadcast (7813eb50): Serailized Json message broadcast on the world rabbit broker by a GNode
        actor
      - RabbitGwSerial (9d70744e): GwSerial protocol message sent on the world rabbit broker
      - MqttJsonBroadcast (16594e91): Serialized Json message following MQTT topic format, sent on the world
        rabbit broker
      - RestApiPost (5b2c69cb): REST API post
      - RestApiPostResponse (7cbba191): REST API post response
      - RestApiGet (167236a6): REST API GET
    """

    Unknown = auto()
    RabbitJsonDirect = auto()
    RabbitJsonBroadcast = auto()
    RabbitGwSerial = auto()
    MqttJsonBroadcast = auto()
    RestApiPost = auto()
    RestApiPostResponse = auto()
    RestApiGet = auto()

    @classmethod
    def default(cls) -> "MessageCategory":
        """
        Returns default value (in this case Unknown)
        """
        return cls.Unknown

    @classmethod
    def values(cls) -> List[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]

    @classmethod
    def version(cls, value: Optional[str] = None) -> str:
        """
        Returns the version of the class (default) used by this package or the
        version of a candidate enum value (always less than or equal to the version
        of the class)

        Args:
            value (Optional[str]): None (for version of the Enum itself) or
            the candidate enum value.

        Raises:
            ValueError: If the value is not one of the enum values.

        Returns:
            str: The version of the enum used by this code (if given no
            value) OR the earliest version of the enum containing the value.
        """
        if value is None:
            return "000"
        if not isinstance(value, str):
            raise ValueError(f"This method applies to strings, not enums")
        if value not in value_to_version.keys():
            raise ValueError(f"Unknown enum value: {value}")
        return value_to_version[value]

    @classmethod
    def enum_name(cls) -> str:
        """
        The name in the GridWorks Type Registry (message.category)
        """
        return "message.category"

    @classmethod
    def enum_version(cls) -> str:
        """
        The version in the GridWorks Type Registry (000)
        """
        return "000"

    @classmethod
    def symbol_to_value(cls, symbol: str) -> str:
        """
        Given the symbol sent in a serialized message, returns the encoded enum.

        Args:
            symbol (str): The candidate symbol.

        Returns:
            str: The encoded value associated to that symbol. If the symbol is not
            recognized - which could happen if the actor making the symbol is using
            a later version of this enum, returns the default value of "Unknown".
        """
        if symbol not in symbol_to_value.keys():
            return cls.default().value
        return symbol_to_value[symbol]

    @classmethod
    def value_to_symbol(cls, value: str) -> str:
        """
        Provides the encoding symbol for a MessageCategory enum to send in seriliazed messages.

        Args:
            symbol (str): The candidate value.

        Returns:
            str: The symbol encoding that value. If the value is not recognized -
            which could happen if the actor making the message used a later version
            of this enum than the actor decoding the message, returns the default
            symbol of "00000000".
        """
        if value not in value_to_symbol.keys():
            return value_to_symbol[cls.default().value]
        return value_to_symbol[value]

    @classmethod
    def symbols(cls) -> List[str]:
        """
        Returns a list of the enum symbols
        """
        return [
            "00000000",
            "9057ab24",
            "7813eb50",
            "9d70744e",
            "16594e91",
            "5b2c69cb",
            "7cbba191",
            "167236a6",
        ]


symbol_to_value = {
    "00000000": "Unknown",
    "9057ab24": "RabbitJsonDirect",
    "7813eb50": "RabbitJsonBroadcast",
    "9d70744e": "RabbitGwSerial",
    "16594e91": "MqttJsonBroadcast",
    "5b2c69cb": "RestApiPost",
    "7cbba191": "RestApiPostResponse",
    "167236a6": "RestApiGet",
}

value_to_symbol = {value: key for key, value in symbol_to_value.items()}

value_to_version = {
    "Unknown": "000",
    "RabbitJsonDirect": "000",
    "RabbitJsonBroadcast": "000",
    "RabbitGwSerial": "000",
    "MqttJsonBroadcast": "000",
    "RestApiPost": "000",
    "RestApiPostResponse": "000",
    "RestApiGet": "000",
}



# Conversion factors

POUNDS_OF_WATER_PER_GALLON = 8.34
BTU_PER_GALLON_PER_DEG_F = POUNDS_OF_WATER_PER_GALLON
BTU_PER_KWH = 3412
KWH_TH_PER_GALLON_PER_DEG_F = BTU_PER_GALLON_PER_DEG_F / BTU_PER_KWH
MINUTES_PER_HOUR = 60
BTU_PER_GALLON_OF_OIL = 140000



from typing import Dict

from gw.enums import MarketPriceUnit
from gw.enums import MarketQuantityUnit
from gw.enums import MarketTypeName
from gw.enums import RecognizedCurrencyUnit


class MarketType:
    by_id: Dict[MarketTypeName, "MarketType"] = {}

    def __new__(cls, name: MarketTypeName, *args, **kwargs) -> "MarketType":  # type: ignore
        try:
            return cls.by_id[name]
        except KeyError:
            instance = super().__new__(cls)
            cls.by_id[name] = instance
            return instance

    def __init__(
        self,
        name: MarketTypeName,
        duration_minutes: int,
        gate_closing_seconds: int,
        price_max: int,
        price_unit: MarketPriceUnit = MarketPriceUnit.USDPerMWh,
        quantity_unit: MarketQuantityUnit = MarketQuantityUnit.AvgMW,
        currency_unit: RecognizedCurrencyUnit = RecognizedCurrencyUnit.USD,
    ):
        self.name = name
        self.duration_minutes = duration_minutes
        self.gate_closing_seconds = gate_closing_seconds
        self.price_max = price_max
        self.price_unit = price_unit
        self.quantity_unit = quantity_unit
        self.currency_unit = currency_unit

    def __repr__(self) -> str:
        s = f"MarketType {self.name}: duration {self.duration_minutes} m, gate closing {self.gate_closing_seconds} s, {self.price_unit.value}, {self.quantity_unit.value}"
        return s


Rt5Gate5 = MarketType(
    name=MarketTypeName.rt5gate5,
    duration_minutes=5,
    gate_closing_seconds=5 * 60,
    price_max=10_000,
)

Rt15Gate5 = MarketType(
    name=MarketTypeName.rt15gate5,
    duration_minutes=15,
    gate_closing_seconds=5 * 60,
    price_max=10_000,
)

Rt30Gate5 = MarketType(
    name=MarketTypeName.rt30gate5,
    duration_minutes=30,
    gate_closing_seconds=5 * 60,
    price_max=10_000,
)

Rt60Gate5 = MarketType(
    name=MarketTypeName.rt60gate5,
    duration_minutes=60,
    gate_closing_seconds=5 * 60,
    price_max=10_000,
)

Rt60Gate30 = MarketType(
    name=MarketTypeName.rt60gate30,
    duration_minutes=60,
    gate_closing_seconds=30 * 60,
    price_max=10_000,
)

Rt60Gate30B = MarketType(
    name=MarketTypeName.rt60gate30b,
    duration_minutes=60,
    gate_closing_seconds=60,
    price_max=10_000,
    quantity_unit=MarketQuantityUnit.AvgkW,
)

Da60 = MarketType(
    name=MarketTypeName.da60,
    duration_minutes=60,
    gate_closing_seconds=1440,
    price_max=10_000,
)



import datetime
import logging
import re
import time
from typing import Any
from typing import Optional
from typing import no_type_check

from pydantic import BaseModel

from gw.enums import MessageCategory
from gw.enums import MessageCategorySymbol


DEFAULT_STEP_DURATION = 0.1

LOG_FORMAT = (
    "%(levelname) -10s %(asctime)s %(name) -30s %(funcName) "
    "-35s %(lineno) -5d: %(message)s"
)
LOGGER = logging.getLogger(__name__)


class RestfulResponse(BaseModel):
    Note: str
    HttpStatusCode: int = 200
    PayloadTypeName: Optional[str] = None
    PayloadAsDict: Optional[Any] = None


snake_add_underscore_to_camel_pattern = re.compile(r"(?<!^)(?=[A-Z])")


def camel_to_snake(name: str) -> str:
    return snake_add_underscore_to_camel_pattern.sub("_", name).lower()


def snake_to_camel(word: str) -> str:
    return "".join(x.capitalize() or "_" for x in word.split("_"))


def message_category_from_symbol(symbol: MessageCategorySymbol) -> MessageCategory:
    category = MessageCategory.Unknown
    if symbol == MessageCategorySymbol.rj:
        category = MessageCategory.RabbitJsonDirect
    elif symbol == MessageCategorySymbol.rjb:
        category = MessageCategory.RabbitJsonBroadcast
    elif symbol == MessageCategorySymbol.s:
        category = MessageCategory.RabbitGwSerial
    elif symbol == MessageCategorySymbol.gw:
        category = MessageCategory.MqttJsonBroadcast
    elif symbol == MessageCategorySymbol.post:
        category = MessageCategory.RestApiPost
    elif symbol == MessageCategorySymbol.postack:
        category = MessageCategory.RestApiPostResponse
    elif symbol == MessageCategorySymbol.get:
        category = MessageCategory.RestApiGet
    return category


def responsive_sleep(
    obj: Any,
    seconds: float,
    step_duration: float = DEFAULT_STEP_DURATION,
    running_field_name: str = "_main_loop_running",
) -> bool:
    """Sleep in way that is more responsive to thread termination: sleep in step_duration increments up to
    specificed seconds, at after each step checking self._main_loop_running"""
    sleeps = int(seconds / step_duration)
    if sleeps * step_duration != seconds:
        last_sleep = seconds - (sleeps * step_duration)
    else:
        last_sleep = 0
    for _ in range(sleeps):
        if getattr(obj, running_field_name):
            time.sleep(step_duration)
    if getattr(obj, running_field_name) and last_sleep > 0:
        time.sleep(last_sleep)
    return getattr(obj, running_field_name)  # type: ignore[no-any-return]


class MessageSummary:
    """Helper class for formating message summaries message receipt/publication single line summaries."""

    DEFAULT_FORMAT = (
        "{timestamp}  {direction:4s}  {actor_alias:33s}  {broker_flag}  {arrow:2s}  {topic:80s}"
        "  {payload_type}"
    )

    @no_type_check
    @classmethod
    def format(
        cls,
        direction: str,
        actor_alias: str,
        topic: str,
        payload_object: Any = None,
        broker_flag=" ",
        timestamp: Optional[datetime.datetime] = None,
    ) -> str:
        """
        Formats a single line summary of message receipt/publication.

        Args:
            direction: "IN" or "OUT"
            actor_alias: The node alias of the sending or receiving actor.
            topic: The destination or source topic.
            payload_object: The payload of the message.
            broker_flag: "*" for the "gw" broker.
            timestamp: datetime.datetime.now(datetime.timezone.utc) by default.

        Returns:
            Formatted string.
        """
        try:
            if timestamp is None:
                timestamp = datetime.datetime.now(datetime.timezone.utc)
            direction = direction[:3].strip().upper()
            if direction in ["OUT", "SND"]:
                arrow = "->"
            elif direction.startswith("IN") or direction.startswith("RCV"):
                arrow = "<-"
            else:
                arrow = "? "
            if hasattr(payload_object, "__class__"):
                payload_str = payload_object.__class__.__name__
            else:
                payload_str = type(payload_object)
            return cls.DEFAULT_FORMAT.format(
                timestamp=timestamp.isoformat(),
                direction=direction,
                actor_alias=actor_alias,
                broker_flag=broker_flag,
                arrow=arrow,
                topic=f"[{topic}]",
                payload_type=payload_str,
            )
        except Exception as e:
            print(f"ouch got {e}")
            return ""
